using System.Collections.Generic;
using RPGM.Core;
using RPGM.UI;
using UnityEngine;

namespace RPGM.Gameplay
{
    /// <summary>
    /// This class implements quests.
    /// </summary>
    public class Quest : MonoBehaviour
    {
        public enum SpawnMode
        {
            CloneAndEnable,
            CloneOnly
        }

        [System.Serializable]
        public class ItemRequirement
        {
            public string item_name;
            public int count = 1;
        }

        public string title;
        public string desc;
        public ConversationScript questInProgressConversation, questCompletedConversation;

        public SpawnMode spawnMode = SpawnMode.CloneAndEnable;
        bool disableItemsOnStart = true;

        [HideInInspector] public GameObject[] enableOnQuestStart;
        [HideInInspector] public GameObject[] spawnOnQuestStart;
        public ItemRequirement[] requiredItems;
        [HideInInspector] public GameObject[] spawnOnQuestComplete;
        [HideInInspector] public InventoryItem[] rewardItems;

        public bool destroySpawnsOnQuestComplete = true;

        public Cutscene introCutscenePrefab, outroCutscenePrefab;

        List<GameObject> cleanup = new List<GameObject>();

        //private bool isStarted = false;
        //private bool isFinished = false;

        private Dictionary<NPCController, bool> isStartedPerNpc = new Dictionary<NPCController, bool>();
        private Dictionary<NPCController, bool> isFinishedPerNpc = new Dictionary<NPCController, bool>();
        private Dictionary<NPCController, int> questEnableTime = new Dictionary<NPCController, int>();

        GameModel model = Schedule.GetModel<GameModel>();

        // Extensions:
        public int tip = 10;  // Amount of Money supervisor receive.
        public int satisfaction_delta = 10;  // On how much satisfcation level of World will change.
        public int satisfaction_delta_npc = 10;  // On how much satisfcation level of NPC will change.

        public int required_time = 10;  // How much time repair will take.

        public int waiting_time_min = 120;  // How much time NPC by default can wait.

        // Dirty hack, to be removed:
        [HideInInspector] public int times_conversation = 0;

        public AudioClip doing_clip;
        public AudioClip done_clip;

        void Awake()
        {
            if (questInProgressConversation == null)
                questInProgressConversation = gameObject.AddComponent<ConversationScript>();
            questInProgressConversation = Utils.CreateQuestConversation(this, questInProgressConversation);
            if (questCompletedConversation == null)
                questCompletedConversation = gameObject.AddComponent<ConversationScript>();
            questCompletedConversation = Utils.CreateFinalQuestConversation(this, questCompletedConversation);
            //if required, make sure that items that will be enabled by this quest are disabled
            if (disableItemsOnStart)
            {
                if (enableOnQuestStart != null)
                    foreach (var i in enableOnQuestStart)
                        if (i != null)
                            i.SetActive(false);

                switch (spawnMode)
                {
                    case SpawnMode.CloneAndEnable:
                        foreach (var i in spawnOnQuestStart)
                        {
                            i.SetActive(false);
                        }
                        break;
                }
            }
        }

        public float PlayDoingClip() {
            if (doing_clip == null) return 0.0f;
            UserInterfaceAudio.PlayClip(doing_clip);
            return doing_clip.length;
        }

        public void PlayDoneClip() {
            if (done_clip == null) return;
            UserInterfaceAudio.PlayClip(done_clip);
        }

        public bool isStarted(NPCController npc) {
            bool value = false;
            if (isStartedPerNpc.TryGetValue(npc, out value)) return value;
            return false;
        }

        public bool isFinished(NPCController npc) {
            bool value = false;
            if (isFinishedPerNpc.TryGetValue(npc, out value)) return value;
            return false;
        }

        public void SetStarted(NPCController npc) {
            isStartedPerNpc[npc] = true;
        }

        public void SetEnabled(NPCController npc) {
            questEnableTime[npc] = Mathf.FloorToInt(model.world_time_min);
        }

        public void SetNotStarted(NPCController npc) {
            isStartedPerNpc[npc] = false;
        }

        public void SetFinished(NPCController npc) {
            isFinishedPerNpc[npc] = true;
        }

        public void SetNotFinished(NPCController npc) {
            isFinishedPerNpc[npc] = false;
        }

        public int ExpiresInMin(NPCController npc) {
            int expiration_time_min = questEnableTime[npc] + waiting_time_min;
            return Mathf.Max(0, expiration_time_min - Mathf.FloorToInt(model.world_time_min));
        }

        public void SetFailed(NPCController npc) {
            // TODO
            // Debug.Log("Set failed for called for npc" + npc.GetInstanceID());
            PunishPlayer(npc);
            SetNotFinished(npc);
            SetNotStarted(npc);
            npc.DisableQuest();
            // Debug.Log("In quest set failed: " + isStarted(npc));            
        }

        public void OnStartQuest(NPCController npc)
        {
            // isFinished = false;
            // isStarted = true;
            if (introCutscenePrefab != null)
            {
                var cs = Instantiate(introCutscenePrefab);
                if (cs.audioClip != null)
                {
                    cs.OnFinish += (i) => model.musicController.CrossFade(model.musicController.audioClip);
                }
            }
            if (enableOnQuestStart != null)
                foreach (var i in enableOnQuestStart)
                    if (i != null)
                        i.SetActive(true);
            switch (spawnMode)
            {
                case SpawnMode.CloneAndEnable:
                    foreach (var i in spawnOnQuestStart)
                    {
                        var clone = GameObject.Instantiate(i);
                        clone.SetActive(true);
                        if (destroySpawnsOnQuestComplete) cleanup.Add(clone);
                    }
                    break;
                case SpawnMode.CloneOnly:
                    foreach (var i in spawnOnQuestStart)
                    {
                        var clone = GameObject.Instantiate(i);
                        if (destroySpawnsOnQuestComplete) cleanup.Add(clone);
                    }
                    break;
            }

        }

        public bool IsQuestComplete(NPCController npc)
        {
            var inv = new HashSet<string>(model.InventoryItems);
            /*foreach (var i in requiredItems)
            {
                if (inv.Contains(i.item.name) && model.GetInventoryCount(i.item.name) >= i.count) continue;
                return false;
            }*/
            // Dirty hack: to be removed.
            // Debug.Log("In quest complete: " + isStarted(npc));            
            return isFinished(npc);
        }

        public bool IsToolsAvailable()
        {
            foreach (var i in requiredItems)
            {
                if (model.GetInventoryCount(i.item_name) >= i.count) continue;
                return false;
            }
            return true;
        }

        public List<string> MissingTools()
        {
            var result = new List<string>();
            foreach (var i in requiredItems)
            {
                if (model.GetInventoryCount(i.item_name) >= i.count) continue;
                Debug.Log(i.item_name + " " + model.GetInventoryCount(i.item_name));
                result.Add(i.item_name);
            }
            return result;
        }

        // public void RewardPlayer(NPCController npc) {
        //     foreach (var i in requiredItems)
        //     {
        //         model.RemoveInventoryItem(i.item, i.count);
        //     }
        //     model.ChangeMoney(tip);
        //     model.ChangeSatisfactionLevel(satisfaction_delta);
        //     model.AdvanceGameTime(required_time);
        //     npc.ChangeSatisfactionLevel(satisfaction_delta_npc);
        //     model.ShowGameState();            
        // }

        public void PunishPlayer(NPCController npc) {
            model.ChangeSatisfactionLevel(-1 * satisfaction_delta);
            npc.ChangeSatisfactionLevel(-1 * satisfaction_delta_npc);
            model.ShowGameState();
        }

        public void RewardItemsToPlayer()
        {
            /*
            foreach (var i in rewardItems)
            {
                MessageBar.Show($"You collected: {i.name} x {i.count}");
                model.AddInventoryItem(i);
                UserInterfaceAudio.OnCollect();
                i.gameObject.SetActive(false);
            }
            if (outroCutscenePrefab != null)
            {
                var cs = Instantiate(outroCutscenePrefab);
                if (cs.audioClip != null)
                {
                    cs.OnFinish += (i) => model.musicController.CrossFade(model.musicController.audioClip);
                }
            }
            */
        }

        public void OnFinishQuest(NPCController npc)
        {
            if (destroySpawnsOnQuestComplete)
            {
                foreach (var i in cleanup)
                {
                    if (i != null) Destroy(i);
                }
            }

            foreach (var i in spawnOnQuestComplete)
            {
                var clone = GameObject.Instantiate(i);
                clone.SetActive(true);
            }
            model.ChangeMoney(tip);
            model.ChangeSatisfactionLevel(satisfaction_delta);
            if (false) {
                // This is not required when timer supports speed up.
                model.AdvanceGameTime(required_time);
            }
            model.ShowGameState();
            SetNotFinished(npc);
            SetNotStarted(npc);
            npc.DisableQuest();
        }

        public ConversationScript getQuestInProgressConversation()
        {
            questInProgressConversation.OnAfterDeserialize();
            return questInProgressConversation;
        }

        public ConversationScript getQuestCompletedConversation()
        {
            questInProgressConversation.OnAfterDeserialize();
            return questCompletedConversation;
        }

    }
}